#include <pch.h>

int main(int argc, char** argv) {
    InitSystems();

    SDL_GPUDevice* device = DeviceCreate();
    SDL_Window* window = SDL_CreateWindow("Brimstone", 800, 600, SDL_WINDOW_RESIZABLE);
    LogAssertSDL(window != NULL, "Failed to create window");
    bool success = SDL_ClaimWindowForGPUDevice(device, window);
    LogAssertSDL(success, "Failed to claim window for device");

    Vertex vertices[] = {
        { { -0.5f, 0.5f, 0.0f }, { 0.0f, 0.0f } },   // top left
        { { -0.5f, -0.5f, 0.0f }, { 0.0f, 1.0f } },  // bottom left
        { { 0.5f, -0.5f, 0.0f }, { 1.0f, 1.0f } },   // bottom right
        { { 0.5f, 0.5f, 0.0f }, { 1.0f, 0.0f } },    // top right
    };
    int indices[] = { 0, 1, 2, 2, 3, 0 };

    SDL_GPUBuffer* vertex_buffer = SDL_CreateGPUBuffer(device,
        &(SDL_GPUBufferCreateInfo){
            .usage = SDL_GPU_BUFFERUSAGE_VERTEX,
            .size = sizeof(vertices),
        });
    LogAssertSDL(vertex_buffer != NULL, "Failed to create GPU buffer");
    UploadToBuffer(device, vertices, sizeof(vertices), vertex_buffer);

    SDL_GPUBuffer* index_buffer = SDL_CreateGPUBuffer(device,
        &(SDL_GPUBufferCreateInfo){
            .usage = SDL_GPU_BUFFERUSAGE_INDEX,
            .size = sizeof(indices),
        });
    LogAssertSDL(index_buffer != NULL, "Failed to create GPU buffer");
    UploadToBuffer(device, indices, sizeof(indices), index_buffer);

    SDL_GPUShader* vertex_shader = ShaderLoad(
        device, "shaders/basic.vert.spv", SDL_GPU_SHADERSTAGE_VERTEX, 1, 0);

    SDL_GPUShader* fragment_shader = ShaderLoad(
        device, "shaders/basic.frag.spv", SDL_GPU_SHADERSTAGE_FRAGMENT, 0, 1);

    SDL_GPUGraphicsPipeline* pipeline = GraphicsPipelineCreate(
        device, window, vertex_shader, fragment_shader);

    ShaderFree(device, vertex_shader);
    ShaderFree(device, fragment_shader);

    success = IMG_Init(IMG_INIT_JPG);
    LogAssertSDL(success, "Failed to initialize SDL_image");

    const char* path = "image.jpg";
    SDL_Surface* image = IMG_Load(path);
    LogAssertSDLf(image != NULL, "Failed to load image: %s", path);

    printf("%d %d\n", image->w, image->h);

    if(image->format != SDL_PIXELFORMAT_ABGR8888) {
        SDL_Surface* converted = SDL_ConvertSurface(image, SDL_PIXELFORMAT_ABGR8888);
        SDL_DestroySurface(image);
        image = converted;
    }
    SDL_GPUTexture* texture = SDL_CreateGPUTexture(device,
        &(SDL_GPUTextureCreateInfo){
            .type = SDL_GPU_TEXTURETYPE_2D,
            .format = SDL_GPU_TEXTUREFORMAT_R8G8B8A8_UNORM,
            .width = image->w,
            .height = image->h,
            .layer_count_or_depth = 1,
            .num_levels = 1,
            .usage = SDL_GPU_TEXTUREUSAGE_SAMPLER,
        });
    LogAssertSDL(texture != NULL, "Failed to create texture");
    UploadToTexture(device, image, texture);
    SDL_GPUSampler* sampler = SDL_CreateGPUSampler(device, &(SDL_GPUSamplerCreateInfo){ 0 });
    LogAssertSDL(sampler != NULL, "Failed to create sampler");

    struct {
        mat4 viewproj;
        mat4 model;
    } ubo;

    bool running = true;
    SDL_Event event;
    while(running) {
        while(SDL_PollEvent(&event)) {
            switch(event.type) {
            case SDL_EVENT_QUIT:
                running = false;
                break;
            }
        }

        int width, height;
        SDL_GetWindowSize(window, &width, &height);

        mat4 view = GLM_MAT4_IDENTITY_INIT;
        mat4 projection = GLM_MAT4_IDENTITY_INIT;

        glm_perspective(glm_rad(70.0f), (float)width / (float)height, 0.1f, 1000.0f, projection);
        glm_translate(view, (vec3){ 0.0f, 0.0f, -2.0f });
        glm_mat4_mul(projection, view, ubo.viewproj);

        SDL_GPUCommandBuffer* cmd = CommandBufferAcquire(device);

        SDL_GPUTexture* swapchain;
        uint32_t swapchain_width, swapchain_height;
        success = SDL_AcquireGPUSwapchainTexture(
            cmd, window, &swapchain, &swapchain_width, &swapchain_height);
        LogAssertSDL(success, "Failed to acquire swapchain texture");
        if(swapchain == NULL) {
            SDL_CancelGPUCommandBuffer(cmd);
            continue;
        }

        glm_mat4_identity(ubo.model);
        glm_rotate(
            ubo.model, glm_rad(sinf(SDL_GetTicks() / 200.0f) * 50.0f), (vec3){ 0.5f, 1.0f, 0.2f });

        SDL_PushGPUVertexUniformData(cmd, 0, &ubo, sizeof(ubo));

        SDL_GPURenderPass* render_pass = SDL_BeginGPURenderPass(cmd,
            &(SDL_GPUColorTargetInfo){
                .texture = swapchain,
                .load_op = SDL_GPU_LOADOP_CLEAR,
                .clear_color = (SDL_FColor){ 0.1f, 0.3f, 0.4f, 1.0f },
            },
            1, NULL);
        LogAssertSDL(render_pass != NULL, "Failed to begin render pass");

        GraphicsPipelineBind(render_pass, pipeline);
        SDL_BindGPUVertexBuffers(render_pass, 0,
            (SDL_GPUBufferBinding[]){
                {
                    .buffer = vertex_buffer,
                    .offset = 0,
                },
            },
            1);
        SDL_BindGPUIndexBuffer(render_pass,
            &(SDL_GPUBufferBinding){
                .buffer = index_buffer,
                .offset = 0,
            },
            1);
        SDL_BindGPUFragmentSamplers(render_pass, 0,
            &(SDL_GPUTextureSamplerBinding){
                .texture = texture,
                .sampler = sampler,
            },
            1);
        SDL_DrawGPUIndexedPrimitives(render_pass, sizeof(indices), 1, 0, 0, 0);
        SDL_EndGPURenderPass(render_pass);
        CommandBufferSubmit(cmd);
    }

    GraphicsPipelineFree(device, pipeline);

    return 0;
}
